import os
import pytest

from riopg import create_pool, Connection


async def get_pool():
    return await create_pool(os.environ.get("DB_URL"))


async def get_connection():
    return await Connection.open(os.environ.get("DB_URL"))


async def test_basic_fetch():
    conn = await get_connection()
    async with conn:
        cur = await conn.cursor()
        await cur.execute("SELECT 1;")
        result = await cur.fetchone()
        assert result == (1,)

    assert conn.closed


async def test_query_param():
    conn = await get_connection()
    async with conn:
        cur = await conn.cursor()
        await cur.execute("SELECT %s", (1,))
        result = await cur.fetchone()
        assert result == (1,)


async def test_cursor():
    conn = await get_connection()
    async with conn:
        cur = await conn.cursor()
        async with cur:
            assert not cur.closed
            await cur.execute("""
            DROP TABLE IF EXISTS users;
            
            CREATE TABLE users (
                id INTEGER PRIMARY KEY,
                username TEXT,
                verified BOOLEAN
            );
            INSERT INTO users VALUES (1, 'ffff', true), (2, 'gggg', false), (3, 'eeee', true);
            """)
            await cur.execute("SELECT username, verified FROM users WHERE verified = TRUE ORDER BY "
                              "id;")
            rows = await cur.fetchall()
            assert rows == [("ffff", True), ("eeee", True)]
            await conn.commit()

            await cur.execute("START TRANSACTION;")
            await cur.execute("INSERT INTO users VALUES (4, 'hhhh', true)")
            await conn.reset()
            await cur.execute("SELECT COUNT(*) FROM users;")
            assert (await cur.fetchone()) == (3,)

            await cur.execute("SELECT username FROM users ORDER BY id;")
            await cur.scroll(1)
            assert (await cur.fetchmany(2)) == [("gggg",), ("eeee",)]

            await cur.execute("SELECT * FROM users ORDER BY id;")
            counter = 0
            async for i in cur:
                assert len(i) == 3
                counter += 1

            assert counter == 3


async def test_pool():
    pool = await get_pool()
    async with pool:
        async with pool.acquire() as conn:
            cur = await conn.cursor()
            await cur.execute("SELECT 2 + 2;")
            result = await cur.fetchone()
            assert result == (4,)

        assert len(pool._connections) == 1
        async with pool.acquire() as conn2:
            print(conn2._connection)
            assert conn2 == conn, "Connection was not reused"
            await conn2.close()

        assert len(pool._connections) == 0
        conn = await pool.acquire()
        await pool.release(conn)

    assert pool._connections[0].closed

    with pytest.raises(RuntimeError):
        await pool.acquire()



"""
pytest helpers
"""
import os

import inspect
import multio
import pytest
from functools import wraps


def _wrapper(fn):
    @wraps(fn)
    def inner(*args, **kwargs):
        return multio.run(fn, *args, **kwargs)

    return inner


# copied from trio
@pytest.hookimpl(tryfirst=True)
def pytest_pyfunc_call(pyfuncitem):
    if inspect.iscoroutinefunction(pyfuncitem.obj):
        pyfuncitem.obj = _wrapper(pyfuncitem.obj)


# not copied from trio
@pytest.fixture(scope="session", autouse=True)
def multio_init(*args, **kwargs):
    multio.init(os.environ.get("MULTIO_LIB"))



# This file is part of riopg.
#
# riopg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# riopg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with riopg.  If not, see <http://www.gnu.org/licenses/>.
"""
.. currentmodule:: riopg.pool
"""
import collections
from typing import Callable

import multio

from riopg import connection as md_connection


async def create_pool(dsn: str, pool_size: int = 12, *,
                      connection_factory: 'Callable[[], md_connection.Connection]' = None) \
        -> 'Pool':
    """
    Creates a new :class:`.Pool`.

    :param dsn: The DSN to connect to the database with.
    :param pool_size: The number of connections to hold at any time.
    :param connection_factory: The pool factory callable to use to
    :return: A new :class:`.Pool`.
    """
    pool = Pool(dsn, pool_size, connection_factory=connection_factory)
    return pool


class _PoolConnectionAcquirer:
    """
    A helper class that allows doing ``async with pool.acquire()``.
    """

    def __init__(self, pool: 'Pool'):
        """
        :param pool: The :class:`.Pool` to use.
        """
        self._pool = pool
        self._conn = None

    async def __aenter__(self) -> 'md_connection.Connection':
        self._conn = await self._pool._acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._pool.release(self._conn)
        return False

    def __await__(self):
        return self._pool._acquire().__await__()


class Pool(object):
    """
    Represents a pool of connections.
    """

    def __init__(self, dsn: str, pool_size: int = 12, *,
                 connection_factory: 'Callable[[], md_connection.Connection]' = None):
        self.dsn = dsn
        self._pool_size = pool_size
        self._connection_factory = connection_factory or md_connection.Connection.open

        self._sema = multio.Semaphore(pool_size)
        self._connections = collections.deque()
        self._closed = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        return False

    async def _make_new_connection(self) -> 'md_connection.Connection':
        """
        Makes a new connection.

        :return: A new :class:`.Connection` or subclass of.
        """
        conn = await self._connection_factory(self.dsn)
        return conn

    async def _acquire(self) -> 'md_connection.Connection':
        """
        Acquires a new connection.

        :return: A :class:`.Connection` from the pool.
        """
        # wait for a new connection to be added
        await self._sema.acquire()
        try:
            conn = self._connections.popleft()
        except IndexError:
            conn = await self._make_new_connection()

        return conn

    def acquire(self) -> '_PoolConnectionAcquirer':
        """
        Acquires a connection from the pool. This returns an object that can be used with
        ``async with`` to automatically release it when done.
        """
        if self._closed:
            raise RuntimeError("The pool is closed")

        return _PoolConnectionAcquirer(self)

    async def release(self, conn: 'md_connection.Connection'):
        """
        Releases a connection.

        :param conn: The :class:`.Connection` to release back to the connection pool.
        """
        if conn is None:
            raise ValueError("Connection cannot be none")

        await multio._maybe_await(self._sema.release())

        if conn._connection.closed:
            # thanks a lot
            return

        await conn.reset()
        self._connections.append(conn)

    async def close(self):
        """
        Closes this pool.
        """
        for connection in self._connections:
            await connection.close()

        self._closed = True



# This file is part of riopg.
#
# riopg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# riopg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with riopg.  If not, see <http://www.gnu.org/licenses/>.
"""
.. currentmodule:: riopg.cursor
"""
from functools import partial
from psycopg2._psycopg import cursor
from typing import Any, Dict, List, Sequence, Tuple, Union

from riopg import connection as md_connection


class Cursor(object):
    """
    Wraps a :class:`psycopg2.cursor` object.
    """

    def __init__(self, connection: 'md_connection.Connection'):
        """
        :param connection: The :class:`.Connection` object this cursor was created under.
        """
        self._connection = connection

        #: The underlying cursor object.
        self._cursor = None  # type: cursor

    async def open(self):
        """
        Opens this cursor.

        This is usually called automatically by :meth:`.Connection.open`.
        """
        self._cursor = await self._connection._do_async(self._connection._cursor)

    # catch-all handler
    def __getattr__(self, item):
        original = getattr(self._cursor, item)
        if not callable(original):
            return original

        # wrap in a _do_async
        def wrapper(s, fn):
            def wrapped(*args, **kwargs):
                return s._connection._do_async(fn, *args, **kwargs)

            return wrapped
        return wrapper(self, original)

    async def __aiter__(self):
        while True:
            res = (await self.fetchone())
            if res is None:
                return

            yield res

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        return False

    async def execute(self, sql: str, params: Union[Tuple[Any], Dict[str, Any]] = None) -> None:
        """
        Executes some SQL in this cursor.

        :param sql: The SQL to execute.
        :param params: The parameters to pass to the SQL query.
        """
        return await self._connection._do_async(partial(self._cursor.execute, sql, params))

    async def fetchone(self) -> Sequence[Any]:
        """
        Fetches one result from this cursor.

        :return: A tuple with the results of the previous query.
        """
        return await self._connection._do_async(self._cursor.fetchone)

    async def fetchmany(self, size: int = None) -> List[Sequence[Any]]:
        """
        Fetches many rows from this cursor.

        :param size: The number of rows to fetch.
        :return: A list of tuples with the results of the current query.
        """
        if size is None:
            size = self._cursor.arraysize

        return await self._connection._do_async(self._cursor.fetchmany, size)

    async def fetchall(self) -> List[Sequence[Any]]:
        """
        Fetches all the rows from this cursor.

        :return: A list of tuples with the results of the current query.
        """
        return await self._connection._do_async(self._cursor.fetchall)

    async def scroll(self, value: int, mode: str = 'relative'):
        """
        Scrolls this cursor.

        :param value: The number of rows to scroll.
        :param mode: The scroll mode to perform.
        """
        return await self._connection._do_async(self._cursor.scroll, value, mode)



# This file is part of riopg.
#
# riopg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# riopg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with riopg.  If not, see <http://www.gnu.org/licenses/>.
"""
.. currentmodule:: riopg.connection
"""

import socket

import inspect
import multio
from psycopg2 import OperationalError, connect
from psycopg2._psycopg import connection
from psycopg2.extensions import POLL_ERROR, POLL_OK, POLL_READ, POLL_WRITE

from riopg import cursor as md_cursor


class Connection(object):
    """
    Wraps a :class:`psycopg2.Connection` object, making it work with an async library.

    Do not construct this object manually; use :meth:`.Connection.open` or :meth:`.Pool.acquire`.
    """

    def __init__(self):
        #: The actual connection object being used.
        self._connection = None  # type: connection

        #: The connection socket being used.
        self._sock = None  # type: socket.socket

        #: The current connection lock. This prevents multiple cursors from executing at the same
        #: time.
        self._lock = multio.Lock()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        return False

    # catch-all handler
    def __getattr__(self, item):
        original = getattr(self._connection, item)
        if not callable(original):
            return original

        # wrap in a _do_async
        def wrapper(s, fn):
            def wrapped(*args, **kwargs):
                return s._do_async(fn, *args, **kwargs)

            return wrapped
        return wrapper(self, original)

    @classmethod
    async def open(cls, *args, **kwargs) -> 'Connection':
        """
        Opens a new connection.
        """
        conn = cls()
        await conn._connect(*args, **kwargs)
        return conn

    async def _wait_callback(self):
        """
        The wait callback. This callback is used for polling the psycopg2 sockets, waiting until
        they are ready.
        """
        while True:
            state = self._connection.poll()
            if state == POLL_OK:
                return

            elif state == POLL_READ:
                await multio.asynclib.wait_read(self._sock)

            elif state == POLL_WRITE:
                await multio.asynclib.wait_write(self._sock)

            elif state == POLL_ERROR:
                raise OperationalError("Polling socket returned error")

    async def _do_async(self, fn, *args):
        """
        Performs a psycopg2 action asynchronously, using the wait callback.
        """
        async with self._lock:
            res = fn(*args)
            if inspect.isawaitable(res):
                res = await res

            await self._wait_callback()  # performs any outstanding network read/writes

        return res

    async def _connect(self, dsn: str):
        """
        Connects the psycopg2 connection.

        :param dsn: The DSN to connect with.
        """
        self._connection = connect(dsn)
        # the socket is required for trio to eat
        # TODO: not AF_INET
        self._sock = socket.fromfd(self._connection.fileno(), socket.AF_INET, socket.SOCK_STREAM)
        await self._wait_callback()

    def _cursor(self):
        """
        Internal implementation of acquiring a cursor.
        """
        return self._connection.cursor()

    async def cursor(self) -> 'md_cursor.Cursor':
        """
        Gets a new cursor object.

        :return: A :class:`.cursor.Cursor` object attached to this connection.
        """
        cur = md_cursor.Cursor(self)
        await cur.open()
        return cur

    async def commit(self):
        """
        Commits the current transaction.
        """
        return await self._do_async(self._connection.commit)

    async def rollback(self):
        """
        Rolls back the current transaction.
        """
        return await self._do_async(self._connection.rollback)

    async def reset(self):
        """
        Resets the current transaction.
        """
        return await self._do_async(self._connection.reset)

    async def close(self):
        """
        Closes this connection.
        """
        await self._do_async(self._connection.reset)
        self._connection.close()  # can't do this async - raises an interfaceerror...



import sys
from pathlib import Path

from setuptools import setup

if sys.version_info[0:2] < (3, 6):
    raise RuntimeError("This package requires Python 3.6+.")

setup(
    name="riopg",
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "dirty-tag"
    },
    packages=[
        "riopg",
    ],
    url="https://github.com/Fuyukai/riopg",
    license="LGPLv3",
    author="Laura Dickinson",
    author_email="l@veriny.tf",
    description="A postgresql library for (cu|t)rio",
    long_description=Path(__file__).with_name("README.rst").read_text(encoding="utf-8"),
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[
        "multio>=0.2.0,<0.4.0",
        "psycopg2>=2.7,<2.8",
    ],
    extras_require={
        "tests": [
            "pytest",
            "pytest-cov",
            "codecov",
            "trio>=0.5.0",
            "curio>=0.8"
        ]
    },
    tests_require=[],
    python_requires=">=3.6.0",
)
